package org.embulk.output;

import com.amazonaws.ClientConfiguration;
import com.amazonaws.auth.AWSStaticCredentialsProvider;
import com.amazonaws.auth.BasicAWSCredentials;
import com.amazonaws.client.builder.AwsClientBuilder.EndpointConfiguration;
import com.amazonaws.services.s3.AmazonS3;
import com.amazonaws.services.s3.AmazonS3ClientBuilder;
import com.amazonaws.services.s3.model.CannedAccessControlList;
import com.amazonaws.services.s3.model.PutObjectRequest;
import com.google.common.base.Optional;
import org.embulk.config.*;
import org.embulk.spi.*;
import org.slf4j.Logger;

import java.io.IOException;
import java.io.OutputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.IllegalFormatException;
import java.util.List;
import java.util.Locale;

public class S3CompatibleFileOutputPlugin implements FileOutputPlugin {
    public interface PluginTask extends Task {
        @Config("path_prefix")
        String getPathPrefix();

        @Config("file_ext")
        String getFileNameExtension();

        @Config("sequence_format")
        @ConfigDefault("\".%03d.%02d\"")
        String getSequenceFormat();

        @Config("bucket")
        String getBucket();

        @Config("region")
        @ConfigDefault("\"us-east-1\"")
        String getRegion();

        @Config("endpoint")
        @ConfigDefault("null")
        Optional<String> getEndpoint();

        @Config("access_key_id")
        @ConfigDefault("null")
        Optional<String> getAccessKeyId();

        @Config("secret_access_key")
        @ConfigDefault("null")
        Optional<String> getSecretAccessKey();

        @Config("proxy_host")
        @ConfigDefault("null")
        Optional<String> getProxyHost();

        @Config("proxy_port")
        @ConfigDefault("null")
        Optional<Integer> getProxyPort();

        @Config("tmp_path")
        @ConfigDefault("null")
        Optional<String> getTempPath();

        @Config("tmp_path_prefix")
        @ConfigDefault("\"embulk-output-s3-\"")
        String getTempPathPrefix();

        @Config("canned_acl")
        @ConfigDefault("null")
        Optional<CannedAccessControlList> getCannedAccessControlList();

        @Config("path_style_access")
        @ConfigDefault("false")
        Optional<Boolean> getPathStyleAccess();

        @Config("signer")
        @ConfigDefault("null")
        Optional<String> getSigner();
    }

    public static class S3FileOutput
            implements FileOutput,
            TransactionalFileOutput {
        private final Logger log = Exec.getLogger(S3CompatibleFileOutputPlugin.class);

        private final String bucket;
        private final String pathPrefix;
        private final String sequenceFormat;
        private final String fileNameExtension;
        private final String tempPathPrefix;
        private final Optional<CannedAccessControlList> cannedAccessControlListOptional;

        private int taskIndex;
        private int fileIndex;
        private AmazonS3 client;
        private OutputStream current;
        private Path tempFilePath;
        private String tempPath = null;

        private static AmazonS3 newS3Client(PluginTask task) {
            AmazonS3ClientBuilder clientBuilder = AmazonS3ClientBuilder.standard();

            // TODO: Support more configurations.
            ClientConfiguration config = new ClientConfiguration();

            if (task.getProxyHost().isPresent()) {
                config.setProxyHost(task.getProxyHost().get());
            }

            if (task.getProxyPort().isPresent()) {
                config.setProxyPort(task.getProxyPort().get());
            }

            if (task.getAccessKeyId().isPresent() && task.getSecretAccessKey().isPresent()) {
                BasicAWSCredentials basicAWSCredentials = new BasicAWSCredentials(
                        task.getAccessKeyId().get(), task.getSecretAccessKey().get());
                clientBuilder = clientBuilder.withCredentials(new AWSStaticCredentialsProvider(basicAWSCredentials));
            }

            if (task.getEndpoint().isPresent()) {
                clientBuilder.setEndpointConfiguration(new EndpointConfiguration(task.getEndpoint().get(), task.getRegion()));
            } else {
                clientBuilder.setRegion(task.getRegion());
            }

            if (task.getPathStyleAccess().isPresent() && task.getPathStyleAccess().get()) {
                clientBuilder = clientBuilder.withPathStyleAccessEnabled(true);
            }

            if (task.getSigner().isPresent()) {
                config.setSignerOverride(task.getSigner().get());
            }

            AmazonS3 client = clientBuilder
                    .withClientConfiguration(config)
                    .build();

            return client;
        }

        S3FileOutput(PluginTask task, int taskIndex) {
            this.taskIndex = taskIndex;
            this.client = newS3Client(task);
            this.bucket = task.getBucket();
            this.pathPrefix = task.getPathPrefix();
            this.sequenceFormat = task.getSequenceFormat();
            this.fileNameExtension = task.getFileNameExtension();
            this.tempPathPrefix = task.getTempPathPrefix();
            if (task.getTempPath().isPresent()) {
                this.tempPath = task.getTempPath().get();
            }
            this.cannedAccessControlListOptional = task.getCannedAccessControlList();
        }

        private static Path newTempFile(String tmpDir, String prefix)
                throws IOException {
            if (tmpDir == null) {
                return Files.createTempFile(prefix, null);
            } else {
                return Files.createTempFile(Paths.get(tmpDir), prefix, null);
            }
        }

        private void deleteTempFile() {
            if (tempFilePath == null) {
                return;
            }

            try {
                Files.delete(tempFilePath);
                tempFilePath = null;
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }

        private String buildCurrentKey() {
            String sequence = String.format(sequenceFormat, taskIndex,
                    fileIndex);
            return pathPrefix + sequence + fileNameExtension;
        }

        private void putFile(Path from, String key) {
            PutObjectRequest request = new PutObjectRequest(bucket, key, from.toFile());
            if (cannedAccessControlListOptional.isPresent()) {
                request.withCannedAcl(cannedAccessControlListOptional.get());
            }
            client.putObject(request);
        }

        private void closeCurrent() {
            if (current == null) {
                return;
            }

            try {
                putFile(tempFilePath, buildCurrentKey());
                fileIndex++;
            } finally {
                try {
                    current.close();
                    current = null;
                } catch (IOException e) {
                    throw new RuntimeException(e);
                } finally {
                    deleteTempFile();
                }
            }
        }

        @Override
        public void nextFile() {
            closeCurrent();

            try {
                tempFilePath = newTempFile(tempPath, tempPathPrefix);

                log.info("Writing S3 file '{}'", buildCurrentKey());

                current = Files.newOutputStream(tempFilePath);
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }

        @Override
        public void add(Buffer buffer) {
            if (current == null) {
                throw new IllegalStateException(
                        "nextFile() must be called before poll()");
            }

            try {
                current.write(buffer.array(), buffer.offset(), buffer.limit());
            } catch (IOException ex) {
                deleteTempFile();
                throw new RuntimeException(ex);
            } finally {
                buffer.release();
            }
        }

        @Override
        public void finish() {
            closeCurrent();
        }

        @Override
        public void close() {
            closeCurrent();
        }

        @Override
        public void abort() {
            deleteTempFile();
        }

        @Override
        public TaskReport commit() {
            TaskReport report = Exec.newTaskReport();
            return report;
        }
    }

    private void validateSequenceFormat(PluginTask task) {
        try {
            @SuppressWarnings("unused")
            String dontCare = String.format(Locale.ENGLISH,
                    task.getSequenceFormat(), 0, 0);
        } catch (IllegalFormatException ex) {
            throw new ConfigException(
                    "Invalid sequence_format: parameter for file output plugin",
                    ex);
        }
    }

    @Override
    public ConfigDiff transaction(ConfigSource config, int taskCount,
                                  Control control) {
        PluginTask task = config.loadConfig(PluginTask.class);

        validateSequenceFormat(task);

        return resume(task.dump(), taskCount, control);
    }

    @Override
    public ConfigDiff resume(TaskSource taskSource, int taskCount,
                             Control control) {
        control.run(taskSource);
        return Exec.newConfigDiff();
    }

    @Override
    public void cleanup(TaskSource taskSource, int taskCount,
                        List<TaskReport> successTaskReports) {
    }

    @Override
    public TransactionalFileOutput open(TaskSource taskSource, int taskIndex) {
        PluginTask task = taskSource.loadTask(PluginTask.class);

        return new S3FileOutput(task, taskIndex);
    }
}
